#include <stdio.h>
#include <stdlib.h>
//#include <string.h>


#define X22 0x00400000 /* vector representation of X^{22} */
#define X11 0x00000800 /* vector representation of X^{11} */
#define MASK12 0xfffff800 /* auxiliary vector for testing */
#define GENPOL 0x00000c75 /* generator polinomial, g(x) */

int bit_pos1[] =
{
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
        
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2,
    0, 0, 1, 1, 2, 2
};

int bit_pos2[] =
{
  23, 11,
  23, 11,
  23, 11,

  22, 10,
  22, 10,
  22, 10,
  
  21, 9,
  21, 9,
  21, 9,
  
  20, 8,
  20, 8,
  20, 8,

  19, 7,
  19, 7,
  19, 7,
  
  18, 6,
  18, 6,
  18, 6,
  
  17, 5,
  17, 5,
  17, 5,
  
  16, 4,
  16, 4,
  16, 4,
  
  15, 3,
  15, 3,
  15, 3,
  
  14, 2,
  14, 2,
  14, 2,
  
  13, 1,
  13, 1,
  13, 1,
  
  12, 0,
  12, 0,
  12, 0
  
};



static long decoding_table[2048];

static int prng[4096];


static void init_prng(void)
{
  int i;

  for (i=0; i < 4096; i++)
  {
    int mask = 0x800000;
    int j;
    int pr;

    prng[i] = 0;
    pr = i << 4;

    for (j=0; j < 24; j++)
    {
      pr = ((173 * pr) + 13849) & 0xFFFF;

      if ((pr & 0x8000) != 0)
      {
	prng[i] |= mask;
      }

      mask = mask >> 1;
    }
  }

}


long get_syndrome(pattern)
/*
* Compute the syndrome corresponding to the given pattern, i.e., the
* remainder after dividing the pattern (when considering it as the vector
* representation of a polynomial) by the generator polynomial, GENPOL.
* In the program this pattern has several meanings: (1) pattern = infomation
* bits, when constructing the encoding table; (2) pattern = error pattern,
* when constructing the decoding table; and (3) pattern = received vector, to
* obtain its syndrome in decoding.
*/
long pattern;
{
// long aux = X22, aux2;
    long aux = X22;
 
    if (pattern >= X11)
       while (pattern & MASK12) {
           while (!(aux & pattern))
              aux = aux >> 1;
           pattern ^= (aux/X11) * GENPOL;
           }
    return(pattern);
}


static int golay2412 (int data, int * decoded)
{
  int block = (data >> 1) & 0x07fffff;
  int corrected_block = block ^ decoding_table[get_syndrome(block)];

  int errs = 0;
  int parity_corr = 0;
  int i;

  for (i = 0; i < 23; i++)
  {
    int mask = 1 << i;
    
    int bit_rcvd = block & mask;
    int bit_corr = corrected_block & mask;
    
    if (bit_corr != 0)
    {
      parity_corr ++;
    }
    
    if (bit_rcvd != bit_corr)
    {
      errs ++;
    }
  }
  
  if ((parity_corr & 0x01) != (data & 0x01))
  {
      errs ++;
  }

  *decoded = corrected_block >> 11;

  return errs;
}


int dstar_dv_decode (const unsigned char * d, int data[3])
{
  int bits[3];
  int i;
  int errs;

  for (i=0; i < 3; i++)
  {
    bits[i] = 0;
  }

  for (i=0; i < 72; i++)
  {
    bits[ bit_pos1[i] ] |= (d[ i >> 3 ] & (0x80 >> (i & 0x07))) ? (1 << bit_pos2[i]) : 0;
  }

  errs = golay2412( bits[0], data );

  errs += golay2412( bits[1] ^ prng[ data[0] & 0x0fff ], data + 1 );

  data[2] = bits[2];

  return errs;
}

int main(int argc, char *argv[])
{

	u_char	d[9], i;
	int errs;
	const char dtmf_chars[16] = "147*2580369#ABCD";

	while( 1 )
	{
		printf("Input: ");
		scanf("%hx%hx%hx%hx%hx%hx%hx%hx%hx", &d[0], &d[1], &d[2], &d[3], &d[4], &d[5], &d[6], &d[7], &d[8] );

		int data[3];
		errs = dstar_dv_decode(d, data);

		if (errs) 
		{
			printf("Error...\n\n");
		} 
//		{
			int x = (data[0] & 0x0ffc);
			int dtmf = (data[0] & 0x03) | ((data[2] & 0x60) >> 3);
			printf("DTMF: %c\n\n", dtmf_chars[dtmf] );
//		}
	}


}
